args = ...
argv = args["argv"]

local function starts_with(str, start)
   return str:sub(1, #start) == start
end

zfs_do = zfs.sync
exceptions = {}
filesystems = {}
snapname = nil
dryrun = false
keep = nil
keepprefix = "@@INVALID@@"
for _, arg in pairs(argv) do
  if starts_with(arg, "--except=") then
    table.insert(exceptions, arg:sub(10, #arg))
  elseif starts_with(arg, "--dry-run") then
    zfs_do = zfs.check
  elseif starts_with(arg, "--keep=") then
    keep = tonumber(arg:sub(8, #arg))
  elseif starts_with(arg, "--keepprefix=") then
    keepprefix = arg:sub(14, #arg)
  elseif starts_with(arg, "-") then
    return {"Invalid argument:", arg}
  elseif starts_with(arg, "@") and #arg > 1 then
    snapname = arg
  else
    table.insert(filesystems, arg)
  end
end

created_snapshots = {}
deleted_snapshots = {}
snapshot_errors = {}
delete_errors = {}

function snapshot_recursive(dataset)
  for _, exception in pairs(exceptions) do
    if starts_with(dataset, exception) then
      return
    end
  end
  for child in zfs.list.children(dataset) do
    snapshot_recursive(child)
  end
  errno, details = zfs_do.snapshot(dataset .. snapname)
  if (errno ~= 0) then
    snapshot_errors[dataset .. snapname] = {}
    snapshot_errors[dataset .. snapname]["errno"] = errno
    snapshot_errors[dataset .. snapname]["details"] = details
  else
    table.insert(created_snapshots, dataset .. snapname)
  end
end

for _, filesystem in pairs(filesystems) do
  snapshot_recursive(filesystem)
end

function delete_recursive(dataset)
  for _, exception in pairs(exceptions) do
    if starts_with(dataset, exception) then
      return
    end
  end
  for child in zfs.list.children(dataset) do
    delete_recursive(child)
  end
  snapshots = {}
  for snapshot in zfs.list.snapshots(dataset) do
    if starts_with(snapshot, dataset .. keepprefix) then
      table.insert(snapshots, snapshot)
    end
  end
  -- Delete earliest snapshots
  table.sort(snapshots)
  if #snapshots > keep then
    for i=1, #snapshots - keep do
      errno, details = zfs_do.destroy(snapshots[i])
      if (errno ~= 0) then
        delete_errors[snapnames[i]] = {}
        delete_errors[snapnames[i]]["errno"] = errno
        delete_errors[snapnames[i]]["details"] = details
      else
        table.insert(deleted_snapshots, snapshots[i])
      end
    end
  end
end

if keep then
  for _, filesystem in pairs(filesystems) do
    delete_recursive(filesystem)
  end
end

result = {}
result["created_snapshots"] = created_snapshots
result["deleted_snapshots"] = deleted_snapshots
result["snapshot_errors"] = snapshot_errors
result["delete_errors"] = delete_errors
return result
